@@ -1,4 +1,4 @@
-/*      $Id: frame.c 29275 2009-01-18 21:45:33Z olivier $
+/*      $Id$
 
         This program is free software; you can redistribute it and/or modify
         it under the terms of the GNU General Public License as published by
@@ -17,7 +17,7 @@
 
 
         oroborus - (c) 2001 Ken Lynch
-        xfwm4    - (c) 2002-2009 Olivier Fourdan
+        xfwm4    - (c) 2002-2015 Olivier Fourdan
 
  */
 
@@ -45,7 +45,6 @@ typedef struct
     xfwmPixmap pm_sides[SIDE_COUNT];
 } FramePixmap;
 
-
 int
 frameDecorationLeft (ScreenInfo *screen_info)
 {
@@ -120,8 +119,7 @@ frameTop (Client * c)
     TRACE ("entering frameTop");
 
     g_return_val_if_fail (c != NULL, 0);
-    if (FLAG_TEST (c->xfwm_flags, XFWM_FLAG_HAS_BORDER)
-        && !FLAG_TEST (c->flags, CLIENT_FLAG_FULLSCREEN))
+    if (CLIENT_HAS_FRAME (c))
     {
         return c->screen_info->title[TITLE_3][ACTIVE].height;
     }
@@ -210,6 +208,112 @@ frameHeight (Client * c)
     return c->height;
 }
 
+int
+frameExtentLeft (Client * c)
+{
+    TRACE ("entering frameExtentLeft");
+
+    g_return_val_if_fail (c != NULL, 0);
+    if (FLAG_TEST (c->flags, CLIENT_FLAG_HAS_FRAME_EXTENTS))
+    {
+        return -c->frame_extents[SIDE_LEFT];
+    }
+    return frameLeft(c);
+}
+
+int
+frameExtentRight (Client * c)
+{
+    TRACE ("entering frameExtentRight");
+
+    g_return_val_if_fail (c != NULL, 0);
+    if (FLAG_TEST (c->flags, CLIENT_FLAG_HAS_FRAME_EXTENTS))
+    {
+        return -c->frame_extents[SIDE_RIGHT];
+    }
+    return frameRight(c);
+}
+
+int
+frameExtentTop (Client * c)
+{
+    TRACE ("entering frameExtentTop");
+
+    g_return_val_if_fail (c != NULL, 0);
+    if (FLAG_TEST (c->flags, CLIENT_FLAG_HAS_FRAME_EXTENTS))
+    {
+        return -c->frame_extents[SIDE_TOP];
+    }
+    return frameTop(c);
+}
+
+int
+frameExtentBottom (Client * c)
+{
+    TRACE ("entering frameExtentBottom");
+
+    g_return_val_if_fail (c != NULL, 0);
+    if (FLAG_TEST (c->flags, CLIENT_FLAG_HAS_FRAME_EXTENTS))
+    {
+        return -c->frame_extents[SIDE_BOTTOM];
+    }
+    return frameBottom(c);
+}
+
+int
+frameExtentX (Client * c)
+{
+    TRACE ("entering frameExtentX");
+
+    g_return_val_if_fail (c != NULL, 0);
+    if (FLAG_TEST (c->flags, CLIENT_FLAG_HAS_FRAME_EXTENTS))
+    {
+        return c->x + c->frame_extents[SIDE_LEFT];
+    }
+    return frameX(c);
+}
+
+int
+frameExtentY (Client * c)
+{
+    TRACE ("entering frameExtentY");
+
+    g_return_val_if_fail (c != NULL, 0);
+    if (FLAG_TEST (c->flags, CLIENT_FLAG_HAS_FRAME_EXTENTS))
+    {
+        return c->y + c->frame_extents[SIDE_TOP];
+    }
+    return frameY(c);
+}
+
+int
+frameExtentWidth (Client * c)
+{
+    TRACE ("entering frameExtentWidth");
+
+    g_return_val_if_fail (c != NULL, 0);
+    if (FLAG_TEST (c->flags, CLIENT_FLAG_HAS_FRAME_EXTENTS))
+    {
+        return MAX (0, c->width - c->frame_extents[SIDE_LEFT]
+                                - c->frame_extents[SIDE_RIGHT]);
+    }
+    return frameWidth(c);
+}
+
+int
+frameExtentHeight (Client * c)
+{
+    TRACE ("entering frameExtentHeight");
+
+    g_return_val_if_fail (c != NULL, 0);
+    if (FLAG_TEST (c->flags, CLIENT_FLAG_HAS_FRAME_EXTENTS))
+    {
+        return MAX (0, c->height - c->frame_extents[SIDE_TOP]
+                                 - c->frame_extents[SIDE_BOTTOM]);
+    }
+    return frameHeight(c);
+}
+
 static int
 frameTopLeftWidth (Client * c, int state)
 {
@@ -281,7 +385,6 @@ static void
 frameCreateTitlePixmap (Client * c, int state, int left, int right, xfwmPixmap * title_pm, xfwmPixmap * top_pm)
 {
     ScreenInfo *screen_info;
-    DisplayInfo *display_info;
     GdkPixmap *gpixmap;
     GdkGCValues values;
     GdkGC *gc;
@@ -298,7 +401,6 @@ frameCreateTitlePixmap (Client * c, int state, int left, int right, xfwmPixmap *
     g_return_if_fail (top_pm);
 
     screen_info = c->screen_info;
-    display_info = screen_info->display_info;
 
     if (left > right)
     {
@@ -313,9 +415,6 @@ frameCreateTitlePixmap (Client * c, int state, int left, int right, xfwmPixmap *
         return;
     }
 
-    layout = gtk_widget_create_pango_layout (myScreenGetGtkWidget (screen_info), c->name);
-    pango_layout_get_pixel_extents (layout, NULL, &logical_rect);
-
     if (left < frameTopLeftWidth (c, state))
     {
         left = frameTopLeftWidth (c, state);
@@ -344,6 +443,10 @@ frameCreateTitlePixmap (Client * c, int state, int left, int right, xfwmPixmap *
         voffset = screen_info->params->title_vertical_offset_inactive;
     }
 
+    layout = gtk_widget_create_pango_layout (myScreenGetGtkWidget (screen_info), c->name);
+    pango_layout_set_auto_dir (layout, FALSE);
+    pango_layout_get_pixel_extents (layout, NULL, &logical_rect);
+
     title_height = screen_info->font_height;
     if (!title_height)
     {
@@ -594,7 +697,6 @@ frameSetShape (Client * c, int state, FramePixmap * frame_pix, int button_x[BUTT
 {
     ScreenInfo *screen_info;
     DisplayInfo *display_info;
-    Window shape_win;
     XRectangle rect;
     xfwmPixmap *my_pixmap;
     int i;
@@ -610,7 +712,14 @@ frameSetShape (Client * c, int state, FramePixmap * frame_pix, int button_x[BUTT
         return;
     }
 
-    shape_win = XCreateSimpleWindow (display_info->dpy, c->frame, 0, 0, frameWidth (c), frameHeight (c), 0, 0, 0);
+    if (screen_info->shape_win == None)
+    {
+        screen_info->shape_win = XCreateSimpleWindow (display_info->dpy, screen_info->xroot, 0, 0, frameWidth (c), frameHeight (c), 0, 0, 0);
+    }
+    else
+    {
+        XResizeWindow (display_info->dpy, screen_info->shape_win, frameWidth (c), frameHeight (c));
+    }
 
     if (FLAG_TEST (c->flags, CLIENT_FLAG_SHADED))
     {
@@ -618,7 +727,7 @@ frameSetShape (Client * c, int state, FramePixmap * frame_pix, int button_x[BUTT
         rect.y = 0;
         rect.width  = frameWidth (c);
         rect.height = frameHeight (c);
-        XShapeCombineRectangles (display_info->dpy, shape_win, ShapeBounding, 0, 0, &rect, 1, ShapeSubtract, Unsorted);
+        XShapeCombineRectangles (display_info->dpy, screen_info->shape_win, ShapeBounding, 0, 0, &rect, 1, ShapeSubtract, Unsorted);
     }
     else if (!FLAG_TEST (c->flags, CLIENT_FLAG_HAS_SHAPE))
     {
@@ -626,11 +735,11 @@ frameSetShape (Client * c, int state, FramePixmap * frame_pix, int button_x[BUTT
         rect.y = frameTop (c);
         rect.width  = c->width;
         rect.height = c->height;
-        XShapeCombineRectangles (display_info->dpy, shape_win, ShapeBounding, 0, 0, &rect, 1, ShapeSet, Unsorted);
+        XShapeCombineRectangles (display_info->dpy, screen_info->shape_win, ShapeBounding, 0, 0, &rect, 1, ShapeSet, Unsorted);
     }
     else
     {
-        XShapeCombineShape (display_info->dpy, shape_win, ShapeBounding, frameLeft (c),
+        XShapeCombineShape (display_info->dpy, screen_info->shape_win, ShapeBounding, frameLeft (c),
                             frameTop (c), c->window, ShapeBounding, ShapeSet);
     }
     if (frame_pix)
@@ -750,20 +859,20 @@ frameSetShape (Client * c, int state, FramePixmap * frame_pix, int button_x[BUTT
         {
             if (xfwmWindowVisible (&c->sides[SIDE_LEFT]))
             {
-                XShapeCombineShape (display_info->dpy, shape_win, ShapeBounding, 0, frameTop (c),
+                XShapeCombineShape (display_info->dpy, screen_info->shape_win, ShapeBounding, 0, frameTop (c),
                                     MYWINDOW_XWINDOW (c->sides[SIDE_LEFT]), ShapeBounding, ShapeUnion);
             }
 
             if (xfwmWindowVisible (&c->sides[SIDE_RIGHT]))
             {
-                XShapeCombineShape (display_info->dpy, shape_win, ShapeBounding, frameWidth (c) - frameRight (c), frameTop (c),
+                XShapeCombineShape (display_info->dpy, screen_info->shape_win, ShapeBounding, frameWidth (c) - frameRight (c), frameTop (c),
                                     MYWINDOW_XWINDOW (c->sides[SIDE_RIGHT]), ShapeBounding, ShapeUnion);
             }
         }
 
         if (xfwmWindowVisible (&c->title))
         {
-            XShapeCombineShape (display_info->dpy, shape_win, ShapeBounding,
+            XShapeCombineShape (display_info->dpy, screen_info->shape_win, ShapeBounding,
                                 frameTopLeftWidth (c, state), 0,
                                 MYWINDOW_XWINDOW (c->title), ShapeBounding, ShapeUnion);
         }
@@ -771,13 +880,13 @@ frameSetShape (Client * c, int state, FramePixmap * frame_pix, int button_x[BUTT
         if (xfwmWindowVisible (&c->corners[CORNER_TOP_LEFT]))
         {
 
-            XShapeCombineShape (display_info->dpy, shape_win, ShapeBounding, 0, 0,
+            XShapeCombineShape (display_info->dpy, screen_info->shape_win, ShapeBounding, 0, 0,
                                 MYWINDOW_XWINDOW (c->corners[CORNER_TOP_LEFT]), ShapeBounding, ShapeUnion);
         }
 
         if (xfwmWindowVisible (&c->sides[SIDE_BOTTOM]))
         {
-            XShapeCombineShape (display_info->dpy, shape_win, ShapeBounding,
+            XShapeCombineShape (display_info->dpy, screen_info->shape_win, ShapeBounding,
                                 screen_info->corners[CORNER_BOTTOM_LEFT][state].width,
                                 frameHeight (c) - frameBottom (c),
                                 MYWINDOW_XWINDOW (c->sides[SIDE_BOTTOM]), ShapeBounding, ShapeUnion);
@@ -785,7 +894,7 @@ frameSetShape (Client * c, int state, FramePixmap * frame_pix, int button_x[BUTT
 
         if (xfwmWindowVisible (&c->sides[SIDE_TOP]))
         {
-            XShapeCombineShape (display_info->dpy, shape_win, ShapeBounding,
+            XShapeCombineShape (display_info->dpy, screen_info->shape_win, ShapeBounding,
                                 screen_info->corners[CORNER_BOTTOM_LEFT][state].width,
                                 frameTop (c) - frameBottom (c),
                                 MYWINDOW_XWINDOW (c->sides[SIDE_TOP]), ShapeBounding, ShapeUnion);
@@ -793,14 +902,14 @@ frameSetShape (Client * c, int state, FramePixmap * frame_pix, int button_x[BUTT
 
         if (xfwmWindowVisible (&c->corners[CORNER_BOTTOM_LEFT]))
         {
-            XShapeCombineShape (display_info->dpy, shape_win, ShapeBounding, 0,
+            XShapeCombineShape (display_info->dpy, screen_info->shape_win, ShapeBounding, 0,
                                 frameHeight (c) - screen_info->corners[CORNER_BOTTOM_LEFT][state].height,
                                 MYWINDOW_XWINDOW (c->corners[CORNER_BOTTOM_LEFT]), ShapeBounding, ShapeUnion);
         }
 
         if (xfwmWindowVisible (&c->corners[CORNER_BOTTOM_RIGHT]))
         {
-            XShapeCombineShape (display_info->dpy, shape_win, ShapeBounding,
+            XShapeCombineShape (display_info->dpy, screen_info->shape_win, ShapeBounding,
                                 frameWidth (c) - screen_info->corners[CORNER_BOTTOM_RIGHT][state].width,
                                 frameHeight (c) - screen_info->corners[CORNER_BOTTOM_RIGHT][state].height,
                                 MYWINDOW_XWINDOW (c->corners[CORNER_BOTTOM_RIGHT]), ShapeBounding, ShapeUnion);
@@ -808,7 +917,7 @@ frameSetShape (Client * c, int state, FramePixmap * frame_pix, int button_x[BUTT
 
         if (xfwmWindowVisible (&c->corners[CORNER_TOP_RIGHT]))
         {
-            XShapeCombineShape (display_info->dpy, shape_win, ShapeBounding,
+            XShapeCombineShape (display_info->dpy, screen_info->shape_win, ShapeBounding,
                                 frameWidth (c) - frameTopRightWidth (c, state),
                                 0, MYWINDOW_XWINDOW (c->corners[CORNER_TOP_RIGHT]), ShapeBounding, ShapeUnion);
         }
@@ -817,7 +926,7 @@ frameSetShape (Client * c, int state, FramePixmap * frame_pix, int button_x[BUTT
         {
             if (xfwmWindowVisible (&c->buttons[i]))
             {
-                XShapeCombineShape (display_info->dpy, shape_win, ShapeBounding, button_x[i],
+                XShapeCombineShape (display_info->dpy, screen_info->shape_win, ShapeBounding, button_x[i],
                                     (frameTop (c) - screen_info->buttons[i][state].height + 1) / 2,
                                     MYWINDOW_XWINDOW (c->buttons[i]), ShapeBounding, ShapeUnion);
             }
@@ -827,17 +936,41 @@ frameSetShape (Client * c, int state, FramePixmap * frame_pix, int button_x[BUTT
     rect.y = 0;
     rect.width  = frameWidth (c);
     rect.height = frameHeight (c);
-    XShapeCombineRectangles (display_info->dpy, shape_win, ShapeBounding, 0, 0, &rect, 1, ShapeIntersect, Unsorted);
-    XShapeCombineShape (display_info->dpy, c->frame, ShapeBounding, 0, 0, shape_win, ShapeBounding, ShapeSet);
+    XShapeCombineRectangles (display_info->dpy, screen_info->shape_win, ShapeBounding, 0, 0, &rect, 1, ShapeIntersect, Unsorted);
+    XShapeCombineShape (display_info->dpy, c->frame, ShapeBounding, 0, 0, screen_info->shape_win, ShapeBounding, ShapeSet);
+}
+
+void
+frameSetShapeInput (Client * c)
+{
+    ScreenInfo *screen_info;
+    DisplayInfo *display_info;
+
+    TRACE ("entering frameSetShapeInput");
+    TRACE ("setting shape input for client (0x%lx)", c->window);
+
+    screen_info = c->screen_info;
+    display_info = screen_info->display_info;
+
+    if (!myDisplayHaveShapeInput(display_info))
+    {
+        return;
+    }
+
+    if (screen_info->shape_win == None)
+    {
+        screen_info->shape_win = XCreateSimpleWindow (display_info->dpy, screen_info->xroot, 0, 0, frameWidth (c), frameHeight (c), 0, 0, 0);
+    }
+    else
+    {
+        XResizeWindow (display_info->dpy, screen_info->shape_win, frameWidth (c), frameHeight (c));
+    }
 
     /* Set Input shape when using XShape extension 1.1 and later */
-    if (display_info->shape_version >= 1001)
-    {
-        XShapeCombineShape (display_info->dpy, shape_win, ShapeInput, frameLeft (c), frameTop (c), c->window, ShapeBounding, ShapeSubtract);
-        XShapeCombineShape (display_info->dpy, shape_win, ShapeInput, frameLeft (c), frameTop (c), c->window, ShapeInput, ShapeUnion);
-        XShapeCombineShape (display_info->dpy, c->frame, ShapeInput, 0, 0, shape_win, ShapeInput, ShapeSet);
-    }
-    XDestroyWindow (display_info->dpy, shape_win);
+    XShapeCombineShape(display_info->dpy, screen_info->shape_win, ShapeInput, 0, 0, c->frame, ShapeBounding, ShapeSet);
+    XShapeCombineShape(display_info->dpy, screen_info->shape_win, ShapeInput, frameLeft (c), frameTop (c), c->window, ShapeBounding, ShapeSubtract);
+    XShapeCombineShape(display_info->dpy, screen_info->shape_win, ShapeInput, frameLeft (c), frameTop (c), c->window, ShapeInput, ShapeUnion);
+    XShapeCombineShape(display_info->dpy, c->frame, ShapeInput, 0, 0, screen_info->shape_win, ShapeInput, ShapeSet);
 }
 
 static void
@@ -846,9 +979,10 @@ frameDrawWin (Client * c)
     ScreenInfo *screen_info;
     FramePixmap frame_pix;
     xfwmPixmap *my_pixmap;
-    int state, i, j, x, button, left, right;
-    int top_width, bottom_width, left_height, right_height;
-    int button_x[BUTTON_COUNT];
+    gint state, x, button, left, right;
+    gint top_width, bottom_width, left_height, right_height;
+    gint button_x[BUTTON_COUNT];
+    guint i, j;
     gboolean requires_clearing;
     gboolean width_changed;
     gboolean height_changed;
@@ -924,8 +1058,7 @@ frameDrawWin (Client * c)
         }
     }
 
-    if (FLAG_TEST (c->xfwm_flags, XFWM_FLAG_HAS_BORDER)
-        && !FLAG_TEST (c->flags, CLIENT_FLAG_FULLSCREEN))
+    if (CLIENT_HAS_FRAME (c))
     {
         /* First, hide the buttons that we don't have... */
         for (i = 0; i < BUTTON_COUNT; i++)
@@ -1194,7 +1327,7 @@ update_frame_idle_cb (gpointer data)
     frameDrawWin (c);
     c->frame_timeout_id = 0;
 
-    return FALSE;
+    return (FALSE);
 }
 
 void
